from gersemi.argument_schema import ArgumentSchema
from gersemi.dumper import Dumper
from gersemi.keyword_kind import KeywordFormatter


class AddCustomTarget(Dumper):
    schema = ArgumentSchema(
        front_positional_arguments=["Name"],
        options=["ALL", "VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS"],
        one_value_keywords=[
            "WORKING_DIRECTORY",
            "COMMENT",
            "JOB_POOL",
            "JOB_SERVER_AWARE",
        ],
        multi_value_keywords=["COMMAND", "DEPENDS", "BYPRODUCTS", "SOURCES"],
        keyword_formatters={"COMMAND": KeywordFormatter.CommandLine},
    )

    def positional_arguments(self, tree):
        if len(tree.children) > 1:
            return super()._format_command_line(tree.children)
        return super().positional_arguments(tree)


add_custom_target = {"add_custom_target": {"__impl": AddCustomTarget}}



from dataclasses import dataclass
from gersemi.types import Tree


class Discard:
    pass


@dataclass
class VisitError(Exception):
    data: str
    node: Tree
    exception: Exception


class Transformer_InPlace:
    def _transform_children(self, children):
        for c in children:
            res = self.transform(c) if isinstance(c, Tree) else c

            if res is not Discard:
                yield res

    def transform(self, tree):
        tree.children = list(self._transform_children(tree.children))
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return tree

        try:
            return f(tree.children)
        except Exception as e:
            raise VisitError(tree.data, tree, e) from e



from gersemi.ast_helpers import is_newline
from gersemi.exceptions import ASTMismatch
from gersemi.transformer import Discard, Transformer_InPlace
from gersemi.types import Tree


def is_not_newline(node):
    return not is_newline(node)


def remove_newlines(nodes):
    yield from filter(is_not_newline, nodes)


class DropIrrelevantNodes(Transformer_InPlace):
    def _drop_node(self, _):
        return Discard

    def line_comment(self, children):
        if not children:
            return Tree("line_comment", [])

        comment_content, *_ = children
        return Tree("line_comment", [comment_content.value.rstrip()])

    def command_invocation(self, children):
        command_name, *rest = children
        return Tree("command_invocation", [str(command_name).lower(), *rest])

    def commented_argument(self, children):
        return Tree("commented_argument", list(remove_newlines(children)))

    def non_command_element(self, children):
        if len(children) == 0:
            return Discard
        return Tree("non_command_element", children)

    def custom_command(self, children):
        _, identifier, arguments, _ = children
        return Tree("command_invocation", [str(identifier).lower(), arguments])

    def command_element(self, children):
        if len(children) == 1:
            return children[0]
        return Tree("command_element", children)

    def arguments(self, children):
        return Tree("arguments", set(remove_newlines(children)))

    def _flatten_blocks(self, children):
        for child in children:
            if getattr(child, "data", None) in ("block", "block_body"):
                yield from child.children
            else:
                yield child

    def start(self, children):
        return Tree("start", list(remove_newlines(self._flatten_blocks(children))))

    def block(self, children):
        return Tree("block", list(remove_newlines(self._flatten_blocks(children))))

    def block_body(self, children):
        return Tree("block_body", list(remove_newlines(self._flatten_blocks(children))))

    newline_or_gap = _drop_node


def drop_irrelevant_nodes(tree):
    return DropIrrelevantNodes().transform(tree)


def check_abstract_syntax_trees_equivalence(lhs, rhs):
    plhs = drop_irrelevant_nodes(lhs)
    prhs = drop_irrelevant_nodes(rhs)
    if plhs != prhs:
        raise ASTMismatch


def check_code_equivalence(parser, lhs_parsed, rhs):
    rhs_parsed = parser.parse(rhs)
    check_abstract_syntax_trees_equivalence(lhs_parsed, rhs_parsed)



import gersemi_rust_backend
from gersemi.ast_helpers import is_line_comment_in
from gersemi.configuration import Spaces
from gersemi.dumper import Dumper


class ConditionSyntaxCommandInvocationDumper(Dumper):
    _inhibit_favour_expansion = True

    def unary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        operation, *rest = tree.children
        formatted_operation = self.visit(operation)
        if len(rest) == 0:
            return formatted_operation

        arg, *_ = rest
        if (
            (not is_line_comment_in(operation))
            and isinstance(self.indent_type, Spaces)
            and (len(formatted_operation.strip()) < self.indent_type)
        ):
            return f"{formatted_operation} {self.visit(arg).lstrip()}"

        with self.indented():
            formatted_arg = self.visit(arg)
        return f"{formatted_operation}\n{formatted_arg}"

    def binary_operation(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        lhs, operation, rhs = tree.children
        formatted_lhs = self.visit(lhs)
        with self.indented():
            formatted_operation = self.visit(operation)
            formatted_rhs = self.visit(rhs)
        return f"{formatted_lhs}\n{formatted_operation}\n{formatted_rhs}"

    def _preprocess_arguments(self, arguments):
        return gersemi_rust_backend.condition_syntax_preprocess_arguments(arguments)

    def _split_arguments(self, arguments):
        if len(arguments) < 1:
            return arguments

        head, *tail = arguments
        return [[head], tail]

    def arguments(self, tree):
        return "\n".join(self.visit_children(tree))

    def complex_argument(self, tree):
        arguments, *_ = tree.children
        arguments = self._preprocess_arguments(arguments)
        result = self._try_to_format_into_single_line(
            arguments.children, prefix="(", postfix=")"
        )
        if result is not None:
            return result

        begin = self._indent("(\n")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}{formatted_arguments}\n{end}"


class ConditionSyntaxCommandInvocationDumperWithDedent(
    ConditionSyntaxCommandInvocationDumper
):
    def format_command(self, tree):
        with self.dedented():
            return super().format_command(tree)


condition_syntax_commands = {
    key: {
        "__impl": ConditionSyntaxCommandInvocationDumperWithDedent
        if key in ("else", "elseif")
        else ConditionSyntaxCommandInvocationDumper
    }
    for key in (
        "elseif",
        "else",
        "endif",
        "endwhile",
        "if",
        "while",
    )
}



from gersemi.types import Tree


class Interpreter:
    def visit_children(self, tree):
        return [
            self.visit(child) if isinstance(child, Tree) else child
            for child in tree.children
        ]

    def __default__(self, tree):
        return self.visit_children(tree)

    def visit(self, tree):
        f = getattr(self, tree.data, self.__default__)
        return f(tree)



from dataclasses import dataclass, field
from typing import Any, Iterable, Mapping, Optional, Sequence
from gersemi.immutable import ImmutableDict, make_immutable
from gersemi.keyword_kind import KeywordFormatter, KeywordPreprocessor
from gersemi.keywords import KeywordMatcher, TwoWordKeywordMatcher

Sections = Mapping[KeywordMatcher, "ArgumentSchema"]
Signatures = Mapping[Optional[KeywordMatcher], "ArgumentSchema"]


@dataclass(eq=True, frozen=True)
class ArgumentSchema:  # pylint: disable=too-many-instance-attributes
    front_positional_arguments: Sequence[str] = ()
    back_positional_arguments: Sequence[str] = ()
    options: Iterable[KeywordMatcher] = ()
    one_value_keywords: Iterable[KeywordMatcher] = ()
    multi_value_keywords: Iterable[KeywordMatcher] = ()
    sections: Sections = field(default_factory=ImmutableDict)
    keyword_formatters: Mapping[str, KeywordFormatter] = field(
        default_factory=ImmutableDict
    )
    keyword_preprocessors: Mapping[str, KeywordPreprocessor] = field(
        default_factory=ImmutableDict
    )


def argument_schema_from_dict(schema: Mapping) -> ArgumentSchema:
    return ArgumentSchema(
        front_positional_arguments=tuple(schema.get("front_positional_arguments", ())),
        back_positional_arguments=tuple(schema.get("back_positional_arguments", ())),
        options=tuple(schema.get("options", ())),
        one_value_keywords=tuple(schema.get("one_value_keywords", ())),
        multi_value_keywords=tuple(schema.get("multi_value_keywords", ())),
        sections=argument_schemas_from_dict(schema.get("sections", {})),
        keyword_formatters=make_immutable(schema.get("keyword_formatters", {})),
        keyword_preprocessors=make_immutable(schema.get("keyword_preprocessors", {})),
    )


def argument_schemas_from_dict(schemas: Mapping) -> Mapping[Any, ArgumentSchema]:
    return make_immutable(
        {key: argument_schema_from_dict(schema) for key, schema in schemas.items()}
    )


def create_schema_patch(source_schema, old_class):
    class Impl(old_class):
        schema = source_schema

    return Impl


@dataclass(eq=True, frozen=True)
class StandardCommand:
    schema: ArgumentSchema
    signatures: Signatures = field(default_factory=ImmutableDict)
    block_end: Optional[str] = None
    canonical_name: Optional[str] = None
    inhibit_favour_expansion: bool = False
    two_words_keywords: Iterable[TwoWordKeywordMatcher] = ()


@dataclass(eq=True, frozen=True)
class SpecializedCommand:
    impl: object
    canonical_name: Optional[str] = None



from typing import Any, Callable
from gersemi.interpreter import Interpreter
from gersemi.types import Node, Token, Tree


def is_tree(tree_type: str) -> Callable[[Node], bool]:
    return lambda element: isinstance(element, Tree) and element.data == tree_type


def is_token(token_type: str) -> Callable[[Node], bool]:
    return lambda element: getattr(element, "type", None) == token_type


is_bracket_comment = is_tree("bracket_comment")
is_commented_argument = is_tree("commented_argument")
is_line_comment = is_tree("line_comment")
is_multi_value_argument = is_tree("multi_value_argument")
is_one_value_argument = is_tree("one_value_argument")
is_option_argument = is_tree("option_argument")
is_positional_arguments = is_tree("positional_arguments")
is_quoted_argument = is_tree("quoted_argument")
is_section = is_tree("section")
is_unquoted_argument = is_tree("unquoted_argument")
is_pair = is_tree("pair")

is_newline = is_token("NEWLINE")


def is_comment(element):
    return is_bracket_comment(element) or is_line_comment(element)


class ContainsLineComment(Interpreter):
    def __default__(self, _) -> bool:
        return False

    def line_comment(self, _) -> bool:
        return True

    def _visit(self, tree: Tree) -> bool:
        subtrees = filter(lambda node: isinstance(node, Tree), tree.children)
        return any(map(self.visit, subtrees))

    arguments = _visit
    commented_argument = _visit
    unary_operation = _visit
    binary_operation = _visit


def is_line_comment_in(node) -> bool:
    return ContainsLineComment().visit(node)


def is_line_comment_in_any_of(nodes) -> bool:
    return any(isinstance(node, Tree) and is_line_comment_in(node) for node in nodes)


def is_keyword(keyword, node):
    if not isinstance(node, Tree):
        return False

    if node.data == "commented_argument":
        return node.children and is_keyword(keyword, node.children[0])

    return node.children and node.children[0] == keyword


def get_value(node, default: Any):
    if isinstance(node, (str, Token)):
        return str(node)

    if node.children:
        if is_commented_argument(node):
            return get_value(node.children[0], default)

        if is_quoted_argument(node):
            return node.children[0].value[1:-1]  # type: ignore

        return node.children[0]

    return default



from gersemi.argument_schema import ArgumentSchema
from gersemi.dumper import Dumper


class SetProperty(Dumper):
    schema = ArgumentSchema(
        options=["GLOBAL", "APPEND", "APPEND_STRING"],
        multi_value_keywords=[
            "TARGET",
            "SOURCE",
            "INSTALL",
            "TEST",
            "CACHE",
            "PROPERTY",
            "TARGET_DIRECTORIES",
            "DIRECTORY",
            "FILE_SET",
        ],
    )
    _keyword_formatters = {"PROPERTY": "_format_property"}

    def _format_property(self, args):
        result = self._try_to_format_into_single_line(args, prefix="", postfix="")
        if result is not None:
            return result

        name, *rest = args
        with self.indented():
            formatted_rest = "\n".join(map(self.visit, rest))
        return f"{self.visit(name)}\n{formatted_rest}"


set_property = {"set_property": {"__impl": SetProperty}}



import argparse
from collections import ChainMap, defaultdict
import collections.abc
from functools import partial
from hashlib import sha1
from pathlib import Path
import sys
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
from ignore import WalkBuilder  # pylint: disable=no-name-in-module
from gersemi.builtin_commands import _builtin_commands
from gersemi.cache import create_cache
from gersemi.configuration import (
    Configuration,
    ControlConfiguration,
    LineRanges,
    MaxWorkers,
    NotSupportedKeys,
    OutcomeConfiguration,
    Workers,
    find_closest_dot_gersemirc,
    make_control_configuration,
    make_outcome_configuration,
    max_number_of_workers,
)
from gersemi.configuration_reports import minimal_report, verbose_report
from gersemi.custom_command_definition_finder import (
    find_custom_command_definitions,
    get_just_definitions,
)
from gersemi.extensions import load_definitions_from_extensions
from gersemi.formatted_file import FormattedFile
from gersemi.formatter import Formatter, NullFormatter, create_formatter
from gersemi.keywords import Keywords
from gersemi.mode import Mode, get_mode
from gersemi.parser import Parser
from gersemi.print_config_kind import PrintConfigKind
from gersemi.result import Error, Result, apply, get_error_message
from gersemi.return_codes import FAIL, INTERNAL_ERROR, SUCCESS
from gersemi.task_result import TaskResult
from gersemi.tasks.check_and_show_diff import check_and_show_diff
from gersemi.tasks.check_formatting import check_formatting
from gersemi.tasks.do_nothing import do_nothing
from gersemi.tasks.format_file import format_file
from gersemi.tasks.forward_to_stdout import forward_to_stdout
from gersemi.tasks.rewrite_in_place import rewrite_in_place
from gersemi.tasks.show_diff import show_diff
from gersemi.utils import fromfile, smart_open
from gersemi.warnings import UnknownCommandWarning

CHUNKSIZE = 16
FILE_PATTERNS = ("CMakeLists.txt", "CMakeLists.txt.in", "*.cmake", "*.cmake.in")


print_to_stdout = partial(print, file=sys.stdout, end="")
print_to_stderr = partial(print, file=sys.stderr)


class WarningSink:
    def __init__(self, quiet):
        self.quiet = quiet
        self.at_least_one_warning_issued = False
        self.records = []

    def __call__(self, s: str):
        self.at_least_one_warning_issued = True
        if not self.quiet:
            self.records.append(s)

    def flush(self):
        for s in self.records:
            print_to_stderr(s)


class StatusCode:
    def __init__(self):
        self.value = SUCCESS

    def __iadd__(self, other):
        if isinstance(other, int):
            self.value = max(self.value, other)
            return self

        if isinstance(other, collections.abc.Iterable):
            for item in other:
                self += item

            return self

        raise RuntimeError(f"Invalid type: {type(other)}")


def get_files_from_directory(path: Path, respect_ignore_files: bool) -> Iterable[Path]:
    if not respect_ignore_files:
        for pattern in FILE_PATTERNS:
            yield from path.rglob(pattern)

        return

    for entry in WalkBuilder(path).require_git(False).build():
        path = entry.path().resolve(True)
        if any(path.match(pattern) for pattern in FILE_PATTERNS):
            yield path


def get_files_from_single_path(
    path: Path, respect_ignore_files: bool
) -> Iterable[Path]:
    if path.is_dir():
        yield from get_files_from_directory(path, respect_ignore_files)
        return

    yield path


def get_files(paths: Iterable[Path], respect_ignore_files: bool) -> List[Path]:
    return sorted(
        {
            item.resolve(True) if item != Path("-") else item
            for path in paths
            for item in get_files_from_single_path(path, respect_ignore_files)
        }
    )


def has_custom_command_definition(code: str) -> bool:
    lowercased = code.lower()
    has_function_definition = "function" in lowercased and "endfunction" in lowercased
    has_macro_definition = "macro" in lowercased and "endmacro" in lowercased
    return has_function_definition or has_macro_definition


def find_custom_command_definitions_in_file_impl(filepath: Path) -> Dict[str, Keywords]:
    with smart_open(filepath, "r") as f:
        code = f.read()
    if not has_custom_command_definition(code):
        return {}

    parse_tree = Parser(_builtin_commands).parse(code)
    return find_custom_command_definitions(parse_tree, filepath)


def find_custom_command_definitions_in_file(
    filepath: Path,
) -> Result[Dict[str, Keywords]]:
    return apply(find_custom_command_definitions_in_file_impl, filepath)


def check_conflicting_definitions(definitions, warning_sink: WarningSink):
    for name, info in definitions.items():
        if len(info) > 1:
            warning_sink(f"Warning: conflicting definitions for '{name}':")
            places = sorted(where for _, where in info)
            for index, where in enumerate(places):
                kind = "(used)   " if index == 0 else "(ignored)"
                warning_sink(f"{kind} {where}")


def find_all_custom_command_definitions(
    paths: Iterable[Path],
    pool,
    warning_sink: WarningSink,
    respect_ignore_files: bool,
) -> Dict[str, Keywords]:
    result: Dict = {}

    try:
        files = get_files(
            paths,
            respect_ignore_files=respect_ignore_files,
        )
    except FileNotFoundError as e:
        # pylint: disable=broad-exception-raised
        raise Exception(f"Definition path doesn't exist: {e.filename}") from e

    find = find_custom_command_definitions_in_file

    for defs in pool.imap_unordered(find, files):
        if isinstance(defs, Error):
            warning_sink(get_error_message(defs))
            continue

        for name, info in defs.items():
            if name in result:
                result[name].extend(info)
            else:
                result[name] = info

    check_conflicting_definitions(result, warning_sink)

    return get_just_definitions(result)


def select_task(mode: Mode, configuration: Configuration):
    return {
        Mode.ForwardToStdout: forward_to_stdout,
        Mode.RewriteInPlace: rewrite_in_place,
        Mode.CheckFormatting: check_formatting,
        Mode.ShowDiff: partial(show_diff, configuration.control.color),
        Mode.CheckFormattingAndShowDiff: partial(
            check_and_show_diff, configuration.control.color
        ),
    }[mode]


def run_task(
    path: Path,
    formatter: Union[NullFormatter, Formatter],
    task: Callable[[FormattedFile], TaskResult],
) -> TaskResult:
    formatted_file: Result[FormattedFile] = apply(format_file, path, formatter)
    if isinstance(formatted_file, Error):
        return TaskResult(
            path=path,
            return_code=INTERNAL_ERROR,
            to_stderr=get_error_message(formatted_file),
        )
    return task(formatted_file)


def consume_task_result(
    task_result: TaskResult,
    configuration: Configuration,
    warning_sink: WarningSink,
) -> Tuple[Path, int, bool]:
    if task_result.to_stdout != "":
        print_to_stdout(task_result.to_stdout)

    warnings = (
        [w for w in task_result.warnings if not isinstance(w, UnknownCommandWarning)]
        if not configuration.outcome.warn_about_unknown_commands
        else task_result.warnings
    )

    for warning in warnings:
        warning_sink(warning.get_message(fromfile(task_result.path)))

    if task_result.to_stderr != "":
        warning_sink(task_result.to_stderr)

    return task_result.path, task_result.return_code, (len(warnings) > 0)


class AdaptivePool:
    def __init__(self, workers: Workers):
        self.workers = (
            max_number_of_workers()
            if isinstance(workers, MaxWorkers)
            else max(1, workers)
        )

    def imap_unordered(self, func, iterable):
        processes = min(self.workers, (len(iterable) + CHUNKSIZE - 1) // CHUNKSIZE)
        if processes <= 1:
            yield from map(func, iterable)
            return

        import multiprocessing

        with multiprocessing.Pool(processes=processes) as pool:
            yield from pool.imap_unordered(func, iterable, CHUNKSIZE)


def summarize_configuration(configuration, extension_definitions):
    hasher = sha1()
    hasher.update(repr(configuration).encode("utf-8"))
    hasher.update(repr(extension_definitions).encode("utf-8"))
    return hasher.hexdigest()


def split_files_by_formatting_state(
    cache,
    configuration: OutcomeConfiguration,
    files: Iterable[Path],
):
    extension_definitions = load_definitions_from_extensions(configuration.extensions)
    summary = summarize_configuration(configuration, extension_definitions)
    known_files = cache.get_files(summary)
    already_formatted_files = []
    files_to_format = []
    for f in files:
        if f not in known_files:
            files_to_format.append(f)
        else:
            s = f.stat()
            if (s.st_size, s.st_mtime_ns) != known_files[f]:
                files_to_format.append(f)
            else:
                already_formatted_files.append(f)

    return already_formatted_files, files_to_format


def store_files_in_cache(
    mode: Mode, cache, configuration_summary: str, files: Iterable[Path]
) -> None:
    if mode in [
        Mode.CheckFormatting,
        Mode.CheckFormattingAndShowDiff,
        Mode.RewriteInPlace,
    ]:
        cache.store_files(configuration_summary, files)


def select_task_for_already_formatted_files(mode: Mode):
    return {
        Mode.ForwardToStdout: forward_to_stdout,
    }.get(mode, do_nothing)


def handle_already_formatted_files(
    mode: Mode,
    configuration: Configuration,
    warning_sink: WarningSink,
    already_formatted_files: Iterable[Path],
) -> Iterable[int]:
    task = select_task_for_already_formatted_files(mode)
    formatter = NullFormatter()
    execute = partial(run_task, formatter=formatter, task=task)
    results = [
        consume_task_result(result, configuration, warning_sink)
        for result in map(execute, already_formatted_files)
    ]
    return (code for _, code, _ in results)


def handle_files_to_format(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    mode: Mode,
    configuration: Configuration,
    cache,
    pool,
    warning_sink: WarningSink,
    files_to_format: Iterable[Path],
    lines_to_format: LineRanges,
) -> Iterable[int]:
    custom_command_definitions = find_all_custom_command_definitions(
        set(configuration.outcome.definitions),
        pool,
        warning_sink,
        configuration.control.respect_ignore_files,
    )
    extension_definitions = load_definitions_from_extensions(
        configuration.outcome.extensions
    )

    formatter = create_formatter(
        configuration.outcome,
        ChainMap(custom_command_definitions, extension_definitions),
        lines_to_format,
    )
    task = select_task(mode, configuration)
    execute = partial(run_task, formatter=formatter, task=task)

    results = [
        consume_task_result(result, configuration, warning_sink)
        for result in pool.imap_unordered(execute, files_to_format)
    ]
    store_files_in_cache(
        mode,
        cache,
        summarize_configuration(configuration.outcome, extension_definitions),
        (
            path
            for path, code, has_warnings in results
            if code == SUCCESS and path != Path("-") and (not has_warnings)
        ),
    )
    return (code for _, code, _ in results)


class GetConfiguration:
    def __init__(
        self,
        args: argparse.Namespace,
        control: ControlConfiguration,
        warning_sink: WarningSink,
    ):
        self.args = args
        self.control = control
        self.processed_configuration_files: List[Path] = []
        self.warning_sink = warning_sink

    def _warn(self, item: NotSupportedKeys, text: str):
        self.warning_sink(f"{item.path}: {text}")

    def _inform_about_not_supported_keys(self, item: NotSupportedKeys):
        if item.path in self.processed_configuration_files:
            return

        if item.path is None:
            return

        self.processed_configuration_files.append(item.path)

        if len(item.command_line_only) > 0:
            keys = ", ".join(sorted(item.command_line_only))
            self._warn(
                item, f"these options are supported only through command line: {keys}"
            )

        if len(item.unknown) > 0:
            keys = ", ".join(sorted(item.unknown))
            self._warn(item, f"these options are not supported: {keys}")

    def __call__(self, configuration_file: Optional[Path]) -> Configuration:
        outcome_configuration, not_supported_keys = make_outcome_configuration(
            configuration_file=configuration_file,
            args=self.args,
        )
        self._inform_about_not_supported_keys(not_supported_keys)
        return Configuration(control=self.control, outcome=outcome_configuration)


def split_files_by_configuration_file(
    paths: Iterable[Path], control: ControlConfiguration
):
    if control.configuration_file is not None:
        return {control.configuration_file: paths}

    result = defaultdict(list)
    for path in paths:
        result[find_closest_dot_gersemirc(path)].append(path)

    return result


def print_configuration_report(
    kind: PrintConfigKind,
    buckets: Dict[Optional[Path], Iterable[Path]],
    get_configuration: GetConfiguration,
):
    report = {
        PrintConfigKind.Minimal: lambda conf_file, conf, _: minimal_report(
            conf_file, conf
        ),
        PrintConfigKind.Verbose: verbose_report,
    }.get(kind, lambda *args: None)

    for config_file, target_files in buckets.items():
        config = get_configuration(config_file)
        result = report(config_file, config.outcome, target_files)
        if result is not None:
            print_to_stdout(result)


# pylint: disable=too-many-locals
def run(args: argparse.Namespace):
    control = make_control_configuration(args)

    try:
        requested_files = get_files(
            args.sources,
            respect_ignore_files=control.respect_ignore_files,
        )
    except FileNotFoundError as e:
        # pylint: disable=broad-exception-raised
        raise Exception(f"Source path doesn't exist: {e.filename}") from e

    if args.line_ranges and len(requested_files) > 1:
        # pylint: disable=broad-exception-raised
        raise Exception("Line range formatting available only with one source file")

    mode = get_mode(args)
    warning_sink = WarningSink(control.quiet)
    pool = AdaptivePool(workers=control.workers)

    buckets = split_files_by_configuration_file(requested_files, control)
    get_configuration = GetConfiguration(args, control, warning_sink)
    if mode == Mode.PrintConfig:
        print_configuration_report(args.print_config, buckets, get_configuration)
        return SUCCESS

    enable_cache = control.cache and (not control.line_ranges)
    with create_cache(enable_cache, control.cache_dir) as cache:
        status_code = StatusCode()
        for config_file, files in buckets.items():
            config = get_configuration(config_file)
            if config.outcome.disable_formatting:
                continue

            already_formatted_files, files_to_format = split_files_by_formatting_state(
                cache, config.outcome, files
            )

            status_code += handle_already_formatted_files(
                mode, config, warning_sink, already_formatted_files
            )
            status_code += handle_files_to_format(
                mode,
                config,
                cache,
                pool,
                warning_sink,
                files_to_format,
                control.line_ranges,
            )

        status_code += (
            FAIL
            if (control.warnings_as_errors and warning_sink.at_least_one_warning_issued)
            else SUCCESS
        )
        warning_sink.flush()
        return status_code.value



from collections import defaultdict
from contextlib import contextmanager
from functools import lru_cache
from itertools import filterfalse
import re
from textwrap import indent
from typing import Dict, List, Optional, Sequence, Tuple, Union
import gersemi_rust_backend
from gersemi.argument_schema import (
    ArgumentSchema,
    Signatures,
    StandardCommand,
    create_schema_patch,
)
from gersemi.ast_helpers import (
    get_value,
    is_comment,
    is_commented_argument,
    is_line_comment,
    is_line_comment_in,
    is_line_comment_in_any_of,
    is_multi_value_argument,
    is_one_value_argument,
    is_option_argument,
    is_pair,
    is_positional_arguments,
    is_section,
)
from gersemi.configuration import (
    Indent,
    ListExpansion,
    OutcomeConfiguration,
    SortOrder,
    Spaces,
    Tabs,
)
from gersemi.keyword_kind import kind_to_formatter
from gersemi.keywords import AnyMatcher
from gersemi.types import Nodes, Tree
from gersemi.warnings import FormatterWarnings, UnknownCommandWarning

BRACKET_ARGUMENT_REGEX = r"(\[(?P<equal_signs>(=*))\[(?:[\s\S]+?)\](?P=equal_signs)\])"
QUOTED_ARGUMENT_REGEX = r'("(?:[^\\\"]|\n|(?:\\(?:[^A-Za-z0-9]|[nrt]))|\\\n)*")'
LINE_COMMENT_BEGIN = "#"
BRACKET_COMMENT_BEGIN = "#["


def flat_split(pattern, string):
    m = re.search(pattern, string)
    if m is None:
        return (string,)

    begin, end = m.span()
    return string[:begin], string[begin:end], string[end:]


def split_by_line_comment(string):
    return flat_split(r"\s*" + LINE_COMMENT_BEGIN, string)


def split_by_bracket_arguments(string):
    return flat_split(BRACKET_ARGUMENT_REGEX, string)


def split_by_quoted_arguments(string):
    return re.split(QUOTED_ARGUMENT_REGEX, string)


def split_into_segments(string):
    head, *comment = split_by_line_comment(string)
    line_comment = "".join(comment)
    if '"' in head:
        head += line_comment
        line_comment = ""
    segments = split_by_bracket_arguments(head)
    result = [split_by_quoted_arguments(segment) for segment in segments]
    result += [[line_comment]]
    return [item for segment in result for item in segment if item != ""]


def indent_segment(segment, indent_symbol):
    if segment[:1] in ["[", '"']:
        return segment

    if segment.startswith(" ") or segment.startswith("\t"):
        return segment

    return indent(segment, indent_symbol, lambda s: not s.startswith("\n"))


def safe_indent(string, indent_symbol):
    segments = split_into_segments(string)
    return "".join(indent_segment(segment, indent_symbol) for segment in segments)


def strip_empty_lines_from_edges(s):
    lines = s.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    while lines and not lines[-1].strip():
        lines.pop()
    return "\n".join(lines)


def find_line_comment_begin(s):
    start = 0
    while True:
        index = s.rfind(LINE_COMMENT_BEGIN, start)
        if index == -1:
            return index

        if index == s.rfind(BRACKET_COMMENT_BEGIN, start):
            start = index + 1
        else:
            return index


def ends_with_line_comment(s):
    return find_line_comment_begin(s) != -1


def remove_common_beginning(s, other):
    index = 0
    for left, right in zip(s, other):
        if left != right:
            break
        index += 1

    return s[index:]


def get_indent(indent_type: Indent) -> str:
    if isinstance(indent_type, Tabs):
        return "\t"
    return " " * indent_type


class WontFit(Exception):
    pass


@lru_cache(maxsize=None)
def create_patch(data, old_class):
    if not isinstance(data, StandardCommand):

        class Impl(data.impl, old_class):  # pylint: disable=function-redefined
            _canonical_name = data.canonical_name

        return Impl

    class Impl(old_class):  # pylint: disable=function-redefined
        _canonical_name = data.canonical_name
        _inhibit_favour_expansion = data.inhibit_favour_expansion
        _two_words_keywords = data.two_words_keywords
        schema = data.schema
        signatures = data.signatures

    return Impl


class Dumper:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    schema: ArgumentSchema
    signatures: Signatures = {}

    _inhibit_favour_expansion: bool = False
    _two_words_keywords: Sequence[Tuple[str, Union[str, AnyMatcher]]] = []
    _keyword_formatters: Dict[str, str] = {}
    _canonical_name: Optional[str] = None

    def __init__(self, configuration: OutcomeConfiguration, known_definitions):
        self.width = configuration.line_length
        self.indent_type = configuration.indent
        self._indent_symbol = get_indent(self.indent_type)
        self.indent_level = 0
        self.favour_expansion = False
        self.unknown_commands_used: Dict[str, List[Tuple[int, int]]] = defaultdict(list)
        self.list_expansion = configuration.list_expansion
        self.sort_order = configuration.sort_order
        self.known_definitions = known_definitions

    @contextmanager
    def patched(self, patch):
        if patch is None:
            yield self
            return

        old_class = type(self)
        try:
            self.__class__ = patch
            yield self
        finally:
            self.__class__ = old_class

    def _get_patch(self, raw_command_name):
        command = self.known_definitions.get(raw_command_name.lower(), None)
        if command is None:
            return None

        return create_patch(command, type(self))

    def command_invocation(self, tree):
        command_name, _ = tree.children
        patch = self._get_patch(str(command_name))
        with self.patched(patch):
            return self.format_command(tree)

    def __default__(self, tree: Tree):
        return "".join(self.visit_children(tree))

    def visit(self, tree):
        f = getattr(self, tree.data, self.__default__)
        return f(tree)

    def visit_children(self, tree):
        yield from (
            self.visit(child) if isinstance(child, Tree) else str(child)
            for child in tree.children
        )

    @property
    def indent_symbol(self):
        return self._indent_symbol * self.indent_level

    def _indent(self, text: str):
        return indent(text, self.indent_symbol)

    def _single_line_helper(self, children, offset):
        width = offset
        for c in children:
            if isinstance(c, Tree):
                if is_line_comment_in(c):
                    raise WontFit()

                result = self.visit(c)
            else:
                result = str(c)

            if "\n" in result:
                raise WontFit()

            width += len(result)
            if width > self.width:
                raise WontFit()

            yield result
            width += 1

    def _try_to_format_into_single_line(
        self, children: Nodes, prefix: str = "", postfix: str = ""
    ) -> Optional[str]:
        if self.favour_expansion:
            return None

        try:
            reserved_space = len(prefix) + len(postfix) + len(self.indent_symbol)
            with self.not_indented():
                formatted = " ".join(self._single_line_helper(children, reserved_space))
            return f"{self.indent_symbol}{prefix}{formatted}{postfix}"
        except WontFit:
            return None

    @contextmanager
    def with_indent_level(self, indent_level):
        old_indent_level = self.indent_level
        try:
            self.indent_level = indent_level
            yield self
        finally:
            self.indent_level = old_indent_level

    def dedented(self):
        return self.with_indent_level(self.indent_level - 1)

    def indented(self):
        return self.with_indent_level(self.indent_level + 1)

    def not_indented(self):
        return self.with_indent_level(0)

    @contextmanager
    def select_expansion_strategy(self):
        old = self.favour_expansion
        try:
            self.favour_expansion = self.list_expansion == ListExpansion.FavourExpansion
            yield self
        finally:
            self.favour_expansion = old

    @contextmanager
    def select_inlining_strategy(self):
        old = self.favour_expansion
        try:
            self.favour_expansion = False
            yield self
        finally:
            self.favour_expansion = old

    def _record_unknown_command(self, command):
        self.unknown_commands_used[str(command)].append((command.line, command.column))

    def get_warnings(self) -> FormatterWarnings:
        if len(self.unknown_commands_used) == 0:
            return []

        return [
            UnknownCommandWarning(command_name=name, positions=positions)
            for name, positions in self.unknown_commands_used.items()
        ]

    def _format_keyword_with_pairs(self, args):
        return "\n".join(map(self.visit, gersemi_rust_backend.pair_arguments(args)))

    def start(self, tree):
        result = self.__default__(tree)
        if result.endswith("\n"):
            return result
        return result + "\n"

    def block(self, tree):
        return "\n".join(filter(None, map(self.visit, tree.children)))

    def block_body(self, tree):
        with self.indented():
            return "".join(self.visit_children(tree))

    def command_element(self, tree):
        invocation, *comment = tree.children
        formatted_invocation = self.visit(invocation)
        if len(comment) == 0:
            return formatted_invocation

        with self.not_indented():
            formatted_comment = self.visit(comment[0])

        return f"{formatted_invocation} {formatted_comment}"

    def non_command_element(self, tree):
        return " ".join(self.visit(child) for child in tree.children)

    def line_comment(self, tree):
        return self._indent("".join(map(str, tree.children))).rstrip()

    def standalone_identifier(self, tree):
        return f"{self.indent_symbol}{tree.children[0]}"

    def _default_format_values(self, values) -> str:
        return "\n".join(map(self.visit, values))

    def positional_arguments(self, tree) -> str:
        return "\n".join(map(self.visit, tree.children))

    def _format_non_option(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        keyword, *values = tree.children
        keyword_as_value = get_value(keyword, None)

        can_be_inlined = (not self.favour_expansion) or (
            keyword is not None
            and (not is_pair(tree))
            and (not is_multi_value_argument(tree))
        )
        if can_be_inlined:
            with self.select_inlining_strategy():
                result = self._try_to_format_into_single_line(tree.children)
                if result is not None:
                    return result

        begin = self.visit(keyword)
        if len(values) == 0:
            return begin

        formatter_kind = self._get_formatter(keyword_as_value)
        if formatter_kind is None:
            formatter_kind = self._keyword_formatters.get(
                keyword_as_value, "_default_format_values"
            )

        with self.indented():
            formatter = getattr(self, formatter_kind)
            formatted_values = formatter(values)
        return f"{begin}\n{formatted_values}"

    def option_argument(self, tree):
        return self.visit(tree.children[0])

    def one_value_argument(self, tree):
        return self._format_non_option(tree)

    def pair(self, tree):
        return self._format_non_option(tree)

    def _get_formatter(self, tree):
        return kind_to_formatter(
            self.schema.keyword_formatters.get(get_value(tree, None), None)
        )

    def _get_preprocessor(self, tree):
        return self.schema.keyword_preprocessors.get(get_value(tree, None), None)

    def _preprocess_keyword_values(self, nodes, preprocessor):
        return gersemi_rust_backend.preprocess_keyword_values(
            nodes=nodes,
            preprocessor=preprocessor,
            case_insensitive=self.sort_order == SortOrder.CaseInsensitive,
        )

    def multi_value_argument(self, tree):
        keyword, *values = tree.children
        preprocessor = self._get_preprocessor(keyword)
        if preprocessor is not None:
            tree.children = [
                keyword,
                *self._preprocess_keyword_values(values, preprocessor),
            ]

        return self._format_non_option(tree)

    def section(self, tree):
        header, *rest = tree.children
        preprocessor = self._get_preprocessor(header)
        if preprocessor is not None:
            rest = self._preprocess_keyword_values(rest, preprocessor)

        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        begin = self.visit(header)
        if len(rest) == 0:
            return begin

        with self.indented():
            formatted_values = "\n".join(map(self.visit, rest))
        return f"{begin}\n{formatted_values}"

    def keyword_argument(self, tree):
        return self._format_non_option(tree)

    def format_command_with_short_name(self, begin, arguments, end):
        with self.indented():
            formatted_arguments = self.visit(arguments).lstrip()
        if (
            not is_line_comment_in_any_of(arguments.children)
            and "\n" not in formatted_arguments
        ):
            return "".join([self._indent(begin), formatted_arguments, end])

        return f"{self._indent(begin)}{formatted_arguments}\n{self._indent(end)}"

    def _format_command_with_long_name(self, begin, arguments, end):
        with self.indented():
            formatted_arguments = self.visit(arguments)
        return "\n".join([self._indent(begin), formatted_arguments, self._indent(end)])

    def _split_arguments(self, arguments: Nodes) -> Nodes:
        return gersemi_rust_backend.dumper_split_arguments(
            self.schema,
            arguments,
        )

    def group_size(self, group):
        if is_positional_arguments(group):
            return len(group.children)
        if is_option_argument(group):
            return 0
        if is_one_value_argument(group):
            return len(group.children) - 1
        if is_multi_value_argument(group):
            return len(group.children) - 1
        if is_section(group):
            section_size = len(group.children) - 1
            subarguments_size = max(map(self.group_size, group.children))
            return max(section_size, subarguments_size)

        return 0

    def _inlining_condition(self, arguments):
        groups = self._split_arguments(arguments.children)
        group_sizes = list(map(self.group_size, groups))
        if (
            self.list_expansion == ListExpansion.FavourExpansion
            and not self._inhibit_favour_expansion
        ):
            return all(size < 2 for size in group_sizes)
        return all(size <= 4 for size in group_sizes)

    def format_command_name(self, identifier):
        identifier = str(identifier)
        if self._canonical_name is not None:
            canonical_name = self._canonical_name
        else:
            if "@" in identifier:
                return identifier

            return identifier.lower()

        if canonical_name.strip().lower() != identifier.strip().lower():
            raise RuntimeError

        return canonical_name

    def format_signature(self, tree):
        raw_identifier, arguments = tree.children
        identifier = self.format_command_name(raw_identifier)
        arguments = self._preprocess_arguments(arguments)
        begin = f"{identifier}("
        end = ")"

        result = self._try_to_format_into_single_line(
            arguments.children, prefix=begin, postfix=end
        )
        if result is not None and self._inlining_condition(arguments):
            return result

        with self.select_expansion_strategy():
            if isinstance(self.indent_type, Spaces) and len(begin) == self.indent_type:
                return self.format_command_with_short_name(begin, arguments, end)
            return self._format_command_with_long_name(begin, arguments, end)

    def arguments(self, tree):
        groups = self._split_arguments(tree.children)
        return "\n".join(map(self.visit, filter(None, groups)))

    def commented_argument(self, tree):
        argument, comment, *_ = tree.children
        formatted_argument = self.visit(argument)
        with self.not_indented():
            formatted_comment = self.visit(comment)
        return f"{formatted_argument} {formatted_comment}"

    def complex_argument(self, tree):
        arguments, *_ = tree.children
        if len(arguments.children) <= 4:
            result = self._try_to_format_into_single_line(
                arguments.children, prefix="(", postfix=")"
            )
            if result is not None:
                return result

        begin = self._indent("(\n")
        with self.indented():
            formatted_arguments = self.visit(arguments)
        end = self._indent(")")
        return f"{begin}{formatted_arguments}\n{end}"

    def bracket_comment(self, tree):
        return f"{self.indent_symbol}{''.join(map(str, tree.children))}"

    def bracket_argument(self, tree):
        return f"{self.indent_symbol}{tree.children[0]}"

    def quoted_argument(self, tree):
        return f"{self.indent_symbol}{tree.children[0]}"

    def unquoted_argument(self, tree):
        return f"{self.indent_symbol}{tree.children[0]}"

    def _preprocess_arguments(self, arguments):
        return gersemi_rust_backend.isolate_two_words_keywords(
            self._two_words_keywords, arguments
        )

    def _get_signature(self, keyword):
        for item, signature in self.signatures.items():
            if item is None:
                continue

            if gersemi_rust_backend.is_one_of_keywords([item], keyword):
                return signature

        return self.signatures.get(None, None)

    def format_command(self, tree):
        _, arguments = tree.children
        arguments = self._preprocess_arguments(arguments)
        arguments_only = filterfalse(is_comment, arguments.children)
        signature = None
        for argument in arguments_only:
            signature = self._get_signature(argument)
            if signature is not None:
                signature = create_schema_patch(signature, type(self))
                break

        with self.patched(signature):
            return self.format_signature(tree)

    def _preprocess_content(self, content):
        if content.strip() == "":
            return ""

        begin = "\n" if content.startswith("\n") else ""
        stripped_content = strip_empty_lines_from_edges(content)
        if ends_with_line_comment(stripped_content):
            end = "\n"
            return f"{begin}{stripped_content}{end}"

        end = "\n" if content.endswith("\n") else ""
        return f"{begin}{stripped_content.rstrip()}{end}"

    def custom_command(self, tree):
        indentation, command_name, _, formatted_arguments = tree.children
        self._record_unknown_command(command_name)
        identifier = self.format_command_name(command_name)
        begin = self._indent(f"{identifier}(")
        content = formatted_arguments.children[0]
        if content == "":
            return f"{begin})"

        with self.not_indented():
            result = self._try_to_format_into_single_line(
                formatted_arguments.children, prefix=begin, postfix=")"
            )
        if result is not None:
            return result

        indent_symbol = remove_common_beginning(self.indent_symbol, indentation)
        body = safe_indent(
            self._preprocess_content(content),
            indent_symbol,
        )
        if not content.startswith("\n"):
            body = body.lstrip(indent_symbol)

        if "\n" not in body:
            end = ")"
        elif body.endswith("\n"):
            end = self._indent(")")
        else:
            end = "\n" + self._indent(")")

        return f"{begin}{body}{end}"

    def _should_start_new_line(self, updated_line, last_line):
        return (
            len(updated_line) > self.width
            or "\n" in updated_line
            or last_line.strip().startswith("#")
        )

    def _format_command_line(self, args):
        head, *tail = args
        lines = [self.visit(head)]
        force_next_line = is_commented_argument(head)
        for arg in tail:
            if is_line_comment(arg):
                lines += [self.visit(arg)]
                continue

            with self.not_indented():
                formatted_arg = self.visit(arg)
            updated_line = f"{lines[-1]} {formatted_arg}"
            if force_next_line or self._should_start_new_line(updated_line, lines[-1]):
                force_next_line = is_commented_argument(arg)
                lines += [self.visit(arg)]
            else:
                lines[-1] = updated_line
                if is_commented_argument(arg):
                    force_next_line = True

        return "\n".join(lines)



from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
import sys


def fromfile(path):
    return "<stdin>" if path == Path("-") else str(path)


def tofile(path):
    return "<stdout>" if path == Path("-") else str(path)


class StdinWrapper:
    @staticmethod
    @lru_cache(maxsize=None)
    def read():
        return sys.stdin.read()


def standard_stream_open(mode):
    if mode is None or mode == "" or "r" in mode:
        return StdinWrapper()
    return sys.stdout


@contextmanager
def smart_open(filename, mode, *args, **kwargs):
    if filename == Path("-"):
        yield standard_stream_open(mode)
    else:
        with open(filename, mode, *args, **kwargs, encoding="utf-8") as fh:
            yield fh



from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass
class Tree:
    data: str
    children: list

    def __hash__(self):
        return hash((self.data, tuple(self.children)))


@dataclass
class Token:
    type: str
    value: str
    line: Optional[int] = None
    column: Optional[int] = None
    start_pos: Optional[int] = None
    end_pos: Optional[int] = None

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.type == other.type and self.value == other.value

        return self.value == other

    def __str__(self):
        return self.value

    def __hash__(self):
        return hash(self.value)

    def __iter__(self):
        return iter(self.value)

    def __deepcopy__(self, _memo):
        return Token(
            self.type, self.value, self.line, self.column, self.start_pos, self.end_pos
        )

    def __getattr__(self, attr):
        return getattr(self.value, attr)


Node = Union[Token, Tree]
Nodes = List[Node]



from dataclasses import dataclass
from typing import Tuple, Union


@dataclass
class Hint:
    keyword: str
    kind: str


class AnyMatcher:
    def __eq__(self, other):
        return True

    def __hash__(self):
        return hash(str(self))


TwoWordKeywordMatcher = Tuple[str, Union[str, AnyMatcher]]
KeywordMatcher = Union[str, TwoWordKeywordMatcher]


@dataclass
class Keywords:
    options: Tuple[str, ...] = ()
    one_value_keywords: Tuple[KeywordMatcher, ...] = ()
    multi_value_keywords: Tuple[KeywordMatcher, ...] = ()
    hints: Tuple[Hint, ...] = ()



from collections import ChainMap
from copy import deepcopy
from functools import lru_cache
import re
from typing import List, Tuple
from gersemi.builtin_commands import _builtin_commands
from gersemi.configuration import LineRanges, OutcomeConfiguration
from gersemi.dumper import Dumper
from gersemi.parser import Parser
from gersemi.sanity_checker import check_code_equivalence
from gersemi.warnings import FormatterWarnings

GERSEMI_OFF = "# gersemi: off"
GERSEMI_ON = "# gersemi: on"
DISABLED_FORMATTING_FENCES = {
    GERSEMI_OFF: GERSEMI_ON,
    "# cmake-format: off": "# cmake-format: on",
    "# fmt: off": "# fmt: on",
}
BUG = "#-#-# gersemi: If you see this there is a bug in gersemi, please report it.#-#-#"
LINE_RANGE_FENCE_OFF = f"{GERSEMI_OFF}\n{BUG}\n"
LINE_RANGE_FENCE_ON = f"{BUG}\n{GERSEMI_ON}\n"


@lru_cache(maxsize=None)
def line_range_fence_regex():
    off_pattern = f"[ \t]*{GERSEMI_OFF}\\n{BUG}\\n"
    on_pattern = f"{BUG}\\n[ \t]*{GERSEMI_ON}\\n"
    return re.compile(f"{off_pattern}|{on_pattern}")


def consume_until(iterator, target):
    while True:
        line = next(iterator, None)
        if line is None:
            return None

        if line.strip() == target:
            return line


def reconstruct_disabled_formatting_zones_impl(original, formatted):
    other = iter(original.splitlines(keepends=True))
    active = iter(formatted.splitlines(keepends=True))
    closing_fence = None

    while True:
        line = next(active, None)
        if line is None:
            break

        command = line.strip()

        if closing_fence is None:
            closing_fence = DISABLED_FORMATTING_FENCES.get(command)
            if closing_fence is not None:
                consume_until(other, command)
                other, active = active, other

        if command == closing_fence:
            gersemi_on = consume_until(other, command)
            if gersemi_on is not None:
                line = gersemi_on
            other, active = active, other
            closing_fence = None

        yield line


def reconstruct_disabled_formatting_zones(original, formatted):
    if all(off_pattern not in original for off_pattern in DISABLED_FORMATTING_FENCES):
        return formatted

    return "".join(reconstruct_disabled_formatting_zones_impl(original, formatted))


def add_line_range_fences_impl(lines: List[str], lines_to_format: LineRanges):
    N = len(lines)

    starts = [r.start for r in lines_to_format]
    ends = [r.end for r in lines_to_format]

    if 1 not in starts:
        yield LINE_RANGE_FENCE_OFF

    for line_number, line in enumerate(lines, start=1):
        if (line_number in starts) and (line_number != 1):
            yield LINE_RANGE_FENCE_ON

        yield line

        if (line_number in ends) and (line_number != N):
            yield LINE_RANGE_FENCE_OFF

    if N not in ends:
        yield LINE_RANGE_FENCE_ON


def add_line_range_fences(code: str, lines_to_format: LineRanges) -> str:
    lines = code.splitlines(keepends=True)
    N = len(lines)
    if max(r.end for r in lines_to_format) > N:
        return code

    return "".join(add_line_range_fences_impl(lines, lines_to_format))


def remove_line_range_fences(formatted_code: str) -> str:
    pattern = line_range_fence_regex()
    return pattern.sub("", formatted_code)


class Formatter:
    def __init__(
        self,
        configuration: OutcomeConfiguration,
        known_definitions,
        lines_to_format: LineRanges,
    ):
        self.configuration = configuration
        self.known_definitions = ChainMap(known_definitions, _builtin_commands)
        self.lines_to_format = lines_to_format

    def format(self, code) -> Tuple[str, FormatterWarnings]:
        if self.lines_to_format:
            code = add_line_range_fences(code, self.lines_to_format)

        parser = Parser(self.known_definitions)
        tree = parser.parse(code)
        if not self.configuration.unsafe:
            original = deepcopy(tree)

        dumper = Dumper(self.configuration, self.known_definitions)
        formatted = dumper.visit(tree)
        if not self.configuration.unsafe:
            check_code_equivalence(parser, original, formatted)

        result = reconstruct_disabled_formatting_zones(code, formatted)
        if self.lines_to_format:
            result = remove_line_range_fences(result)

        return result, dumper.get_warnings()


class NullFormatter:
    def format(self, code):
        return code, []


def create_formatter(
    configuration: OutcomeConfiguration,
    known_definitions,
    lines_to_format,
):
    return Formatter(configuration, known_definitions, lines_to_format)



import argparse
from dataclasses import fields
import pathlib
import sys
import gersemi_rust_backend
from gersemi.__version__ import __title__, __version__
from gersemi.configuration import (
    ControlConfiguration,
    ListExpansion,
    OutcomeConfiguration,
    SortOrder,
    indent_type,
    line_ranges,
    normalize_definitions,
    normalize_extensions,
    normalize_path,
    sanitize_list_expansion,
    workers_type,
)
from gersemi.configuration_reports import default_report
from gersemi.print_config_kind import PrintConfigKind, print_config_kind
from gersemi.return_codes import FAIL, SUCCESS
from gersemi.runner import print_to_stderr, run

FROZEN = getattr(sys, "frozen", False)
MISSING = "(missing)"

try:
    from colorama import __version__ as colorama_version

except ImportError:
    colorama_version = MISSING


class ShowVersion(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        frozen_suffix = " (frozen)" if FROZEN else ""
        print(f"{__title__} {__version__}{frozen_suffix}")
        print(f"colorama {colorama_version}")
        print(f"Python {sys.version}")
        print(gersemi_rust_backend.version())
        sys.exit(SUCCESS)


def toggle_action(predicate):
    class Impl(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(
                namespace,
                self.dest,
                not predicate(option_string),
            )

    return Impl


toggle_with_no_prefix = toggle_action(lambda s: s.startswith("--no-"))


def create_argparser():
    parser = argparse.ArgumentParser(
        description="A formatter to make your CMake code the real treasure.",
        prog="gersemi",
        add_help=False,
    )
    modes_group = parser.add_argument_group("modes")
    modes_group.add_argument(
        "-c",
        "--check",
        dest="check_formatting",
        action="store_true",
        help=f"""
    Check if files require reformatting.
    Return {SUCCESS} when there's nothing to reformat.
    Return {FAIL} when some files would be reformatted.
    It can be used together with --diff.
            """,
    )
    modes_group.add_argument(
        "-i",
        "--in-place",
        dest="in_place",
        action="store_true",
        help="Format files in-place.",
    )
    modes_group.add_argument(
        "--diff",
        dest="show_diff",
        action="store_true",
        help="""
    Show diff on stdout for each formatted file instead.
    It can be used together with --check.
        """,
    )
    modes_group.add_argument(
        "--print-config",
        dest="print_config",
        choices=[e.value for e in PrintConfigKind],
        help=f"""Print configuration for files.
        {" ".join(attr.description for attr in PrintConfigKind)}
        Command line arguments are taken into consideration just
        as they would be for formatting.
        When configuration file is found values in "definitions" are printed as relative
        paths, otherwise absolute paths are printed.
        Output can be placed in .gersemirc file verbatim.""",
    )
    modes_group.add_argument(
        "--version",
        nargs=0,
        dest="show_version",
        action=ShowVersion,
        help="Show version.",
    )
    modes_group.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    outcome_conf_doc: dict[str, str] = {
        item.name: item.metadata["description"] for item in fields(OutcomeConfiguration)
    }
    outcome_configuration_group = parser.add_argument_group(
        title="outcome configuration", description=OutcomeConfiguration.__doc__
    )
    outcome_configuration_group.add_argument(
        "-l",
        "--line-length",
        metavar="INTEGER",
        dest="line_length",
        type=int,
        help=f"{outcome_conf_doc['line_length']} [default: {OutcomeConfiguration.line_length}]",
    )
    outcome_configuration_group.add_argument(
        "--indent",
        metavar="(INTEGER | tabs)",
        dest="indent",
        type=indent_type,
        help=f"{outcome_conf_doc['indent']} [default: {repr(OutcomeConfiguration.indent)}]",
    )
    outcome_configuration_group.add_argument(
        "--safe",
        "--unsafe",
        dest="unsafe",
        action=toggle_action(lambda s: s == "--safe"),
        nargs=0,
        default=None,
        help=f"""
    {outcome_conf_doc["unsafe"]}
    [default: skip sanity checks]
            """,
    )
    outcome_configuration_group.add_argument(
        "--definitions",
        dest="definitions",
        metavar="src",
        default=None,
        nargs="+",
        type=pathlib.Path,
        help=outcome_conf_doc["definitions"],
    )
    outcome_configuration_group.add_argument(
        "--list-expansion",
        dest="list_expansion",
        choices=[attr.value for attr in ListExpansion],
        help=f"""
    {outcome_conf_doc["list_expansion"]}
    {" ".join(attr.description for attr in ListExpansion)}
    [default: {OutcomeConfiguration.list_expansion.value}]
            """,
    )
    outcome_configuration_group.add_argument(
        "--warn-about-unknown-commands",
        "--no-warn-about-unknown-commands",
        dest="warn_about_unknown_commands",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
    {outcome_conf_doc["warn_about_unknown_commands"]}
    [default: warnings enabled, same as --warn-about-unknown-commands]
        """,
    )
    outcome_configuration_group.add_argument(
        "--disable-formatting",
        "--enable-formatting",
        dest="disable_formatting",
        action=toggle_action(lambda s: s == "--enable-formatting"),
        nargs=0,
        default=None,
        help=f"""
    {outcome_conf_doc["disable_formatting"]}
    [default: formatting enabled]
        """,
    )
    outcome_configuration_group.add_argument(
        "--extensions",
        dest="extensions",
        metavar="extension-name-or-path",
        default=None,
        nargs="+",
        type=str,
        help=outcome_conf_doc["extensions"],
    )
    outcome_configuration_group.add_argument(
        "--sort-order",
        dest="sort_order",
        choices=[attr.value for attr in SortOrder],
        help=f"""
    {outcome_conf_doc["sort_order"]}
    {" ".join(attr.description for attr in SortOrder)}
    [default: {OutcomeConfiguration.sort_order.value}]
            """,
    )

    control_conf_doc: dict[str, str] = {
        item.name: item.metadata["description"] for item in fields(ControlConfiguration)
    }
    control_configuration_group = parser.add_argument_group(
        title="control configuration", description=ControlConfiguration.__doc__
    )
    control_configuration_group.add_argument(
        "-q",
        "--quiet",
        "--no-quiet",
        dest="quiet",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"{control_conf_doc['quiet']} [default: don't skip, same as --no-quiet]",
    )

    if colorama_version == MISSING:
        warn_about_missing_colorama = " Warning: missing colorama."
    else:
        warn_about_missing_colorama = ""

    control_configuration_group.add_argument(
        "--color",
        "--no-color",
        dest="color",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
        {control_conf_doc["color"]}
        {warn_about_missing_colorama}
        [default: don't colorize diff, same as --no-color]
        """,
    )
    control_configuration_group.add_argument(
        "-w",
        "--workers",
        metavar="(INTEGER | max)",
        dest="workers",
        type=workers_type,
        help=f"""
    {control_conf_doc["workers"]} [default: {repr(ControlConfiguration.workers)}]
        """,
    )
    control_configuration_group.add_argument(
        "--cache",
        "--no-cache",
        dest="cache",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
    {control_conf_doc["cache"]}
    [default: cache enabled, same as --cache]
        """,
    )
    control_configuration_group.add_argument(
        "--cache-dir",
        dest="cache_dir",
        type=pathlib.Path,
        default=None,
        help=f"""
    {control_conf_doc["cache_dir"]}
    [default: omitted]
        """,
    )
    control_configuration_group.add_argument(
        "--config",
        dest="configuration_file",
        type=pathlib.Path,
        default=None,
        help=f"""
    {control_conf_doc["configuration_file"]}
    [default: omitted]
        """,
    )
    control_configuration_group.add_argument(
        "--warnings-as-errors",
        dest="warnings_as_errors",
        action="store_true",
        default=None,
        help=control_conf_doc["warnings_as_errors"],
    )
    control_configuration_group.add_argument(
        "--line-ranges",
        dest="line_ranges",
        default=[],
        type=line_ranges,
        action="append",
        help=control_conf_doc["line_ranges"],
    )
    control_configuration_group.add_argument(
        "--respect-ignore-files",
        "--no-respect-ignore-files",
        dest="respect_ignore_files",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
    {control_conf_doc["respect_ignore_files"]}
    [default: respect ignore files, same as --respect-ignore-files]
        """,
    )

    parser.add_argument(
        dest="sources",
        metavar="src",
        nargs="*",
        type=pathlib.Path,
        help="""
    File or directory to format. When directory is provided then CMakeLists.txt,
    CMakeLists.txt.in and files with .cmake/.cmake.in extension are automatically discovered.
    If only `-` is provided, input is taken from stdin instead.
            """,
    )

    return parser


def is_stdin_mixed_with_file_input(sources):
    if sources is None:
        return False

    return pathlib.Path("-") in sources and len(sources) != 1


def postprocess_args(args):
    args.sources = set(args.sources)
    if args.definitions is not None:
        args.definitions = set(args.definitions)

    args.definitions = normalize_definitions(args.definitions)
    args.extensions = normalize_extensions(args.extensions)
    args.list_expansion = sanitize_list_expansion(args.list_expansion)
    args.print_config = print_config_kind(args.print_config)

    if args.configuration_file is not None:
        args.configuration_file = normalize_path(args.configuration_file)

    args.line_ranges = {line_range for arg in args.line_ranges for line_range in arg}


def error(text):
    print_to_stderr(text)
    sys.exit(FAIL)


def main():
    if FROZEN:
        from multiprocessing import freeze_support

        freeze_support()

    try:
        argparser = create_argparser()
        args = argparser.parse_args()

        if args.print_config == PrintConfigKind.Default.value:
            print(default_report())
            sys.exit(SUCCESS)

        postprocess_args(args)

        if any(map(is_stdin_mixed_with_file_input, [args.sources, args.definitions])):
            error("Don't mix stdin with file input")

        if len(args.sources) < 1:
            sys.exit(SUCCESS)

        sys.exit(run(args))
    except Exception as exception:  # pylint: disable=broad-exception-caught
        error(exception)


if __name__ == "__main__":
    main()



from collections import abc


class ImmutableDict(dict):
    def __setitem__(self, key, value):
        raise TypeError("mutation not allowed")

    def __delitem__(self, key):
        raise TypeError("mutation not allowed")

    def __hash__(self):
        return hash(frozenset(self.items()))

    def __reduce__(self):
        return type(self), (dict(self),)


def make_immutable(thing):
    if isinstance(thing, abc.Mapping):
        return ImmutableDict(
            {key: make_immutable(value) for key, value in thing.items()}
        )

    if isinstance(thing, str):
        return thing

    if isinstance(thing, abc.Collection):
        return tuple(map(make_immutable, thing))

    return thing



import gersemi_rust_backend


class Parser:
    def __init__(self, known_definitions=None):
        self.known_definitions = {} if known_definitions is None else known_definitions

    def parse(self, text):
        return gersemi_rust_backend.parse(
            text,
            blocks=(
                ("if", "endif"),
                ("foreach", "endforeach"),
                ("function", "endfunction"),
                ("macro", "endmacro"),
                ("while", "endwhile"),
                ("block", "endblock"),
                *(
                    (name, definition.block_end)
                    for name, definition in self.known_definitions.items()
                    if getattr(definition, "block_end", None)
                ),
            ),
            known_commands=tuple(key for key in self.known_definitions),
        )
